from splverifier.api.config import Record
from splverifier.integration.featurehouse import get_features_from_expression
import os


sharedstate = Record(
    variability_encode=False,
    selected_features=None,
    actions=dict(),
)

class Writer(object):
    indent_level = 0

    def write(self, *args):
        for arg in args:
            print ' ' * self.indent_level + arg

    def indent(self):
        self.indent_level += 4

    def dedent(self):
        self.indent_level = max(0, self.indent_level - 4)


class FileWriter(Writer):
    
    def __init__(self, filename):
        self.fp = open(filename, 'w')
        
    def write(self, *args):
        for arg in args:
            self.fp.write(' ' * self.indent_level + arg + '\n')
    
    def close(self):
        self.fp.close()
    

class ScenarioDescriptionElement(object):
    pass

class StringWrapper(ScenarioDescriptionElement):
    def __init__(self, expr):
        self.expr = expr
        
    def generate_code(self, writer):
        writer.write(self.expr)
        
    def __repr__(self):
        return self.expr

class Call(StringWrapper):
    
    def generate_code(self, writer):
        if sharedstate.selected_features:
            if self.expr in sharedstate.actions:
                if sharedstate.variability_encode:
                    writer.write('if (__SELECTED_FEATURE_%s)' % (sharedstate.actions[self.expr],))
                else:
                    if not sharedstate.actions[self.expr] in sharedstate.selected_features:
                        return
            
        writer.write(self.expr + '();') 
        
class ExprWrapper(ScenarioDescriptionElement):
    def __init__(self, expr):
        if isinstance(expr, str):
            expr = Call(expr)
        self.expr = expr

class ListWrapper(ScenarioDescriptionElement):
    def __init__(self, *args):
        def wrap(expr):
            if isinstance(expr, str):
                return Call(expr)
            else:
                return expr

        self.elems = [wrap(elem) for elem in args]

class Scenario(ListWrapper):
    inline=True
    def generate_code(self, writer, elems=None):
        if elems is None:
            elems = self.elems

        if self.inline:
            for elem in elems:
                elem.generate_code(writer)
        else:
            raise Exception, 'Not implemented'
        
class C(StringWrapper):
    pass

class Optional(ExprWrapper):
    def generate_code(self, writer):
        writer.write('if (get_nondet()) {')
        writer.indent()
        writer.write(self.expr)
        writer.dedent()
        writer.write('}')
        
class Star(ExprWrapper):
    def generate_code(self, writer):
        writer.write('while (1) {')
        writer.indent()
        writer.write('if (get_nondet()) break;')        
        writer.write(self.expr)
        writer.dedent()
        writer.write('}')
            
class Plus(ExprWrapper):
    def generate_code(self, writer):
        writer.write('while (1) {')
        writer.indent()
        writer.write(self.expr)
        writer.write('if (get_nondet()) break;')
        writer.dedent()
        writer.write('}')

class Permutation(Scenario):
    def generate_code11(self, writer, r=None):
        if r != 0:
            from itertools import permutations
            perms = list(permutations(self.elems, r=r))
            generate_choice(
                writer,
                perms,
                lambda w, i: super(Permutation, self).generate_code(w, elems=i)
            )


uid = 0

def get_uid():
    global uid
    uid += 1
    return uid


class SubsetPermutation(Permutation):
    
    def generate_code(self, writer):
        uids=[]
        writer.write('while (1) {')
        writer.indent()
        for i in self.elems:
            uid = get_uid()
            uids.append(uid)
            writer.write('int op%d = 0;' % (uid,))
        
        for i in range(len(self.elems)):
            uid = uids[i]
            elem = self.elems[i]
            writer.write('if (!op%d && get_nondet()) {' % (uid,))
            writer.indent()
            elem.generate_code(writer)
            writer.write('continue;')
            writer.dedent()
            writer.write('}')
            
        writer.write('if (get_nondet()) break;')
        writer.dedent()
        writer.write('}')
    
    def generate_code11(self, writer):        
        generate_choice(
            writer,
            range(len(self.elems) + 1),
            lambda w, i: super(SubsetPermutation, self).generate_code(w, r=i)
        )


class BaseFeature(object):
    def generate_code(self, writer):
        self.SCENARIO.generate_code(writer)

def generate_choice(writer, choicelist, item_generator):

    writer.write('if (get_nondet()) {')
    writer.indent()
    first = choicelist[0]
    item_generator(writer, first)
    for item in choicelist[1:-1]:
        writer.dedent()
        writer.write('} else if(get_nondet()) {')
        writer.indent()
        item_generator(writer, item)
        
    writer.dedent()
    writer.write('} else {')
    writer.indent()
    last = choicelist[-1]
    item_generator(writer, last)
    writer.dedent()
    writer.write('}')

def get_tree_repr(root):
    if not isinstance(root, ScenarioDescriptionElement):
        raise Exception('Node is not a ScenarioDescriptionElement: ' + root)

    if isinstance(root, ListWrapper):
        return '(%s, %s)' % (
            type(root).__name__,
            ', '.join([get_tree_repr(elem) for elem in root.elems]),
        )
    elif isinstance(root, ExprWrapper):
        return '(%s, %s)' % (
            type(root).__name__,
            get_tree_repr(root.expr),
        )       
    else:
        return '(%s, %s)' % (
            type(root).__name__,
            root.expr,
        )
    

def generate_code(root, writer=Writer()):
    root.generate_code(writer)

def _generate_scenario(scenario, outfile):
    global uid   
    uid = 0

    writer = FileWriter(outfile)
    
    if sharedstate.variability_encode:
        writer.write("""#include "featureselect.h"\n""") 
         
    writer.write('void test() {')
    writer.indent()
    scenario.generate_code(writer)
    writer.dedent()
    writer.write('}')
    writer.close()

def generate_scenario(scenario, outfile, expression=None, selected_features=None, variability_encode=False, action_dict=None):
    if variability_encode:
        sharedstate['selected_features'] = [True]
    else:
        if expression:
            sharedstate['selected_features'] = get_features_from_expression(expression)
        elif selected_features:
            sharedstate['selected_features'] = selected_features
        else:
            raise Exception('Either expression or selected_features needs to be specified.')
            
    sharedstate['actions'] = action_dict
    sharedstate['variability_encode'] = variability_encode
    _generate_scenario(scenario, outfile)
    

